# -*- coding: utf-8 -*-
import sys
import wx
import wx.propgrid as wxpg
from DePyTranslation import Materialien, Gruppe

# ATTRIBUTE DEFINITIONS ========================================================
objectName = ['__class__']
objectAttributes = ['gruppe', 'name', 'position', 'laenge', 'breite', 'hoehe', 'radius',
                    'dicke', 'laengePfeil', 'breitePfeil', 'laengeSpitze',
                    'breiteSpitze', 'achse', 'material', 'farbe',
                    'deckkraft', 'sichtbar']
shownAttributes = objectName + objectAttributes

def filterAttributes(obj):
    allAttributes = filter(lambda aname: not aname.startswith('_'), dir(obj))
    exceptions = ['__class__']
    allAttributes.extend(exceptions)
    return filter(lambda a: a in allAttributes, shownAttributes)

# CUSTOM PROPERTY GRID FRAMEWORK ===============================================
class StaticEditorBase(object):
    def UpdateControl(self, property, ctrl):
        ctrl.SetLabel(property.GetDisplayedString())

    def DrawValue(self, dc, rect, property, text):
        if not property.IsValueUnspecified():
            dc.DrawText(property.GetDisplayedString(), rect.x+5, rect.y)

    def GetValueFromControl(self, property, ctrl):
        """ Return tuple (wasSuccess, newValue), where wasSuccess is True if
            different value was acquired succesfully.
        """
        tc = ctrl
        textVal = tc.GetValue()

        if property.UsesAutoUnspecified() and not textVal:
            return (True, None)

        res, value = property.StringToValue(textVal, wxpg.PG_EDITABLE_VALUE)

        # Changing unspecified always causes event (returning
        # True here should be enough to trigger it).
        if not res and value is None:
            res = True

        return (res, value)

    def SetValueToUnspecified(self, property, ctrl):
        ctrl.Remove(0,len(ctrl.GetValue()))

    def SetControlStringValue(self, property, ctrl, text):
        ctrl.SetLabel(text)

    def OnFocus(self, property, ctrl):
        ctrl.SetSelection(-1,-1)
        ctrl.SetFocus()
        
class StaticEditor(StaticEditorBase,wxpg.PyEditor):
    def __init__(self):
        wxpg.PyEditor.__init__(self)

    def CreateControls(self, propgrid, property, pos, sz):
        """ Create the actual wxPython controls here for editing the
            property value.
            You must use propgrid.GetPanel() as parent for created controls.
            Return value is either single editor control or tuple of two
            editor controls, of which first is the primary one and second
            is usually a button.
        """
        try:
            parent = propgrid.GetPanel()
            x, y = pos
            w, h = sz

            s = property.GetDisplayedString();
            tc = wx.StaticText(parent, wxpg.PG_SUBID1, s, (x+3,y+2), (w-3,h-2))
            tc.SetForegroundColour((0,0,0))
            tc.SetBackgroundColour((255,255,255))
            return tc
        except:
            import traceback
            traceback.print_exc()

    def OnEvent(self, propgrid, property, ctrl, event):
        """ Return True if modified editor value should be committed to
            the property. To just mark the property value modified, call
            propgrid.EditorsValueWasModified().
        """
        return False

PropertyArgs = {}
symbolFont = wx.Font(14, wx.DEFAULT, wx.NORMAL, wx.NORMAL, False, "Webdings")
class SelectObjectEditor(StaticEditorBase,wxpg.PyEditor):
    def __init__(self):
        wxpg.PyEditor.__init__(self)

    def CreateControls(self, propgrid, property, pos, sz):
        """ Create the actual wxPython controls here for editing the
            property value.
            You must use propgrid.GetPanel() as parent for created controls.
            Return value is either single editor control or tuple of two
            editor controls, of which first is the primary one and second
            is usually a button.
        """
        try:
            parent = propgrid.GetPanel()
            x, y = pos
            w, h = sz

            # Make room for button
            bh = h #propgrid.GetRowHeight()
            bw = 1.1 * bh
            w -= bw

            # create controls
            s = property.GetDisplayedString();
            tc = wx.StaticText(parent, wxpg.PG_SUBID1, s, (x+3,y+2), (w-3,h-2))
            tc.SetForegroundColour((0,0,0))
            tc.SetBackgroundColour((255,255,255))

            self.btn = wx.Button(parent, wxpg.PG_SUBID2, u"8", (x+w, y),
                                 (bw, bh), wx.WANTS_CHARS)
            self.btn.SetFont(symbolFont)
            self.btn.SetToolTipString(u"Auswählen")
            
            return (tc,self.btn)
        except:
            import traceback
            traceback.print_exc()

    def OnEvent(self, propgrid, property, ctrl, event):
        """ Return True if modified editor value should be committed to
            the property. To just mark the property value modified, call
            propgrid.EditorsValueWasModified().
        """
        if event.GetEventType() == wx.wxEVT_COMMAND_BUTTON_CLICKED:
            (pg, target) = PropertyArgs[property.GetName()]
            #print "before call"
            wx.CallAfter(pg.SetObject, target)
            #print "after call"
            
        return False

class AdvancedObjectEditor(StaticEditorBase, wxpg.PyEditor):
    def __init__(self):
        wxpg.PyEditor.__init__(self)

    def CreateControls(self, propgrid, property, pos, sz):
        """ Create the actual wxPython controls here for editing the
            property value.
            You must use propgrid.GetPanel() as parent for created controls.
            Return value is either single editor control or tuple of two
            editor controls, of which first is the primary one and second
            is usually a button.
        """
        try:
            parent = propgrid.GetPanel()
            self.parent = parent
            tcWidth = 0.0

            # buttons
            buttons = wxpg.PGMultiButton(propgrid,sz)
            
            buttons.AddButton(u"N")
            btn = buttons.GetButton(0)
            btn.SetFont(symbolFont)
            btn.SetToolTipString(u"Sichtbarkeit ändern")
            w, h = btn.GetSize()
            btn.SetSize((1.2*w,h))
            tcWidth -= 0.2*w
            x, y = btn.GetPosition()
            btn.SetPosition((x-0.2*w,y))
            
            buttons.AddButton(u"8")
            btn = buttons.GetButton(1)
            btn.SetFont(symbolFont)
            btn.SetToolTipString(u"Auswählen")

            # create controls
            x, y = pos
            w, h = buttons.GetPrimarySize()
            s = property.GetDisplayedString()
            tc = wx.StaticText(parent, wxpg.PG_SUBID1, s, (x+4.5,y+2),
                               (w-4.5,h-2))
            self.tc = tc
            self.tc.SetForegroundColour(self.GetTextColor(property))
            tc.SetBackgroundColour((255,255,255))

            # finish button creation
            buttons.Finalize(propgrid, pos)
            self.buttons = buttons
            
            return (tc, buttons)
        except:
            import traceback
            traceback.print_exc()

    def GetTextColor(self, property):
        if PropertyArgs[property.GetName()][1].sichtbar:
            return (0,0,0)
        else:
            return (150,150,150)

    def DrawValue(self, dc, rect, property, text):
        if not property.IsValueUnspecified():
            dc.SetTextForeground(self.GetTextColor(property))
            dc.DrawText(property.GetDisplayedString(), rect.x+5, rect.y)

    def OnEvent(self, propgrid, property, ctrl, event):
        """ Return True if modified editor value should be committed to
            the property. To just mark the property value modified, call
            propgrid.EditorsValueWasModified().
        """
        if event.GetEventType() == wx.wxEVT_COMMAND_BUTTON_CLICKED:
            buttons = self.buttons
            evtId = event.GetId()
            
            if evtId == buttons.GetButtonId(0):
                (pg,target) = PropertyArgs[property.GetName()]
                target.sichtbar = not target.sichtbar

                # update Label Color
                self.tc.SetForegroundColour(self.GetTextColor(property))
                self.tc.SetLabel("") #workaround to initiate repaint
                self.tc.SetLabel(property.GetDisplayedString())
                return False
            elif evtId == buttons.GetButtonId(1):
                (pg,target) = PropertyArgs[property.GetName()]
                #print "before call"
                wx.CallAfter(pg.SetObject,target)
                #print "after call"
                
        return False
        
# ------------------------------ property manager ------------------------------
class AbstractPropManager(object):
    def __init__(self, obj, attribute, property=None):
        self.obj = obj
        self.att = attribute
        
        if not property: property = self.CreateProperty()
        self.prop = property

    def CreateProperty(self):
        raise NotImplementedError("PropertyManager.CreateProperty()")

    def PostInit(self, pg, propID):
        '''Gets called after creation and __init__'''
        pass
        
    def CalcValue(self):
        '''Called when updating the value'''
        return getattr(self.obj,self.att)

    def Process(self, value):
        '''Process a value change'''
        setattr(self.obj,self.att,value)
        return False # don't update property grid
    
    def Update(self):
        '''Check for updates'''
        new = self.CalcValue()
        if self.prop.GetValue() != new: self.prop.SetValue(new)

    def GetProperty(self):
        return self.prop

class FloatPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, property=None):
        AbstractPropManager.__init__(self, obj, attribute, property)
        
    def CreateProperty(self):
        return wxpg.FloatProperty(name=self.att, label=self.att,
                                  value=self.CalcValue())

    def Process(self, value):
        AbstractPropManager.Process(self, value)
        return True # update property grid

class BoolPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, property=None, checkbox=True):
        self.cb = checkbox
        AbstractPropManager.__init__(self, obj, attribute, property)        
        
    def CreateProperty(self):
        return wxpg.BoolProperty(name=self.att, label=self.att,
                                 value=self.CalcValue())
            
    def PostInit(self, pg, propID):
        if self.cb: pg.SetPropertyAttribute(self.prop.GetName(), "UseCheckbox",
                                            True)
    
class StringPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, property=None):
        AbstractPropManager.__init__(self, obj, attribute, property)
        
    def CreateProperty(self):
        return wxpg.StringProperty(name=self.att, label=self.att,
                                   value=self.CalcValue())
    
class StaticPropManager(StringPropManager):
    def __init__(self, obj, attribute, property=None):
        StringPropManager.__init__(self, obj, attribute, property)

    def PostInit(self, pg, propID):
        pg.SetPropertyEditor(propID, "StaticEditor")

    def Process(self, value):
        return False # don't update property grid

class ClassPropManager(StaticPropManager):
    def __init__(self, obj, property=None):
        StaticPropManager.__init__(self, obj, 'Klasse', property)

    def CalcValue(self):
        return self.obj.__class__.__name__

    def PostInit(self, pg, propID):
        pg.DisableProperty(propID)

class GroupPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, property=None):
        AbstractPropManager.__init__(self, obj, attribute, property)

    def CreateProperty(self):
        return wxpg.StringProperty(name=self.att, label=self.att,
                                   value=self.CalcValue())

    def CalcValue(self):
        return AbstractPropManager.CalcValue(self).name

    def PostInit(self, pg, propID):
        pg.SetPropertyEditor(propID, "SelectObjectEditor")
        PropertyArgs[self.att] = (pg, getattr(self.obj, self.att))

    def Process(self, value):
        return False

class ObjectPropManager(AbstractPropManager):
    def __init__(self, obj, property=None):
        AbstractPropManager.__init__(self, obj, "", property)

    def CreateProperty(self):
        return wxpg.StringProperty(name=self.obj.name,
                                   label=self.obj.__class__.__name__,
                                   value=self.CalcValue())

    def CalcValue(self):
        return self.obj.name #__class__.__name__

    def PostInit(self, pg, propID):
        pg.SetPropertyEditor(propID, "AdvancedObjectEditor")
        PropertyArgs[self.obj.name] = (pg, self.obj)

    def Update(self):
        pass

    def Process(self, value):
        return False

class ColourPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, property=None):
        AbstractPropManager.__init__(self, obj, attribute, property)
        
    def CreateProperty(self):
        return wxpg.ColourProperty(name=self.att, label=self.att,
                                   value=self.CalcValue())

    def CalcValue(self):
        vals = tuple(val*255.0 for val in getattr(self.obj, self.att))
        vals = tuple(int(v) for v in vals)
        vals = list(vals)
        vals.append(255)
        return tuple(vals)

    def Process(self,col):
        col.__class__ = wx.Colour
        col = (col.Red()/255.0,col.Green()/255.0,col.Blue()/255.0)
        return AbstractPropManager.Process(self, col)
        
class EnumPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, enum, choices, property=None):
        self.enum = enum
        self.choices = choices
        AbstractPropManager.__init__(self, obj, attribute, property)

    def CreateProperty(self):
        return wxpg.EnumProperty(self.att, self.att, self.choices,
                                 range(len(self.choices)), self.CalcValue())

    def CalcValue(self):
        current = getattr(self.obj, self.att)
        i = 0
        for e in self.choices:
            if current == getattr(self.enum, e):
                return i
            i += 1
        return -1

    def Process(self, currentID):        
        e = getattr(self.enum, self.choices[currentID])
        print 'enum[{}]: {}'.format(currentID, e)
        return AbstractPropManager.Process(self, e)

class ListLengthPropManager(StaticPropManager):
    def __init__(self, name, list, property=None):
        self.name = name
        self.list = list
        StaticPropManager.__init__(self, None, name, property)

    def CreateProperty(self):
        return wxpg.IntProperty(name=self.name, label=self.name,
                                value=self.CalcValue())
        
    def CalcValue(self):
        return len(self.list)

class TupleIndexPropManager(AbstractPropManager):
    def __init__(self, obj, attribute, index, suffix):
        self.index = index
        self.suff = suffix
        AbstractPropManager.__init__(self, obj, attribute, None)

    def CreateProperty(self):
        n = "{} {}".format(self.att, self.suff)
        return wxpg.FloatProperty(name=n, label=n, value=self.CalcValue())
        
    def CalcValue(self):
        return getattr(self.obj, self.att)[self.index]

    def Process(self, value):
        tupl = getattr(self.obj, self.att)
        asList = list(tupl)
        asList[self.index] = value
        tupl = tuple(asList)
        AbstractPropManager.Process(self, tupl)
        return True # update property grid

class TuplePropManager(StaticPropManager):
    def __init__(self, obj, attribute, suffix="XYZ", collaps=False):
        self.suffix = list(suffix)
        self.collaps = collaps
        StaticPropManager.__init__(self, obj, attribute, None)
        
    def CalcValue(self):
        return '<composed>'

    def PostInit(self, pg, propID):
        StaticPropManager.PostInit(self, pg, propID)
        for i in range(len(getattr(self.obj, self.att))):
            man = TupleIndexPropManager(self.obj, self.att, i,
                                        self.suffix[min(i,len(self.suffix)-1)])
            pg.AppendManager(man, propID)
        if self.collaps: pg.Collapse(propID)

    def Update(self):
        pass

        
# ------------------------------------ grid ------------------------------------                  
class ObjectPG(wxpg.PropertyGrid):
    '''Shows and changes attributes of a vPython object '''
    def __init__(self, parent, style=0):
        super(ObjectPG,self).__init__(parent, style)
        self.obj = object()
        self.managerMap = {}
        self.OnObjectChange = (lambda newObj : None)
        self.Bind(wxpg.EVT_PG_CHANGED, self.OnValueChange)

        # NOTE: Editor must be registered *before* adding a property
        # that uses it
        if not getattr(sys, '_PropGridEditorsRegistered', False):
            self.RegisterEditor(StaticEditor)
            self.RegisterEditor(SelectObjectEditor)
            self.RegisterEditor(AdvancedObjectEditor)
            # ensure we only do it once
            sys._PropGridEditorsRegistered = True

    # fill Dictionary
    managerDict = {}
    
    def CreateStringManager(self, attr):
        return StringPropManager(self.obj, attr)
    def CreateGroupManager(self, attr):
        g = getattr(self.obj, attr)
        if g: return GroupPropManager(self.obj, attr)
        return None
    def CreateBoolManager(self, attr):
        return BoolPropManager(self.obj, attr, checkbox=True)
    def CreateFloatManager(self, attr):
        return FloatPropManager(self.obj, attr)
    def CreateColourManager(self, attr):
        return ColourPropManager(self.obj, attr)
    def CreateClassManager(self, attr):
        return ClassPropManager(self.obj)
    def CreateMaterialManager(self, attr):
        return EnumPropManager(self.obj, attr, Materialien,
                               Materialien._GetKeys())
    def CreateVectorManager(self, attr):
        return TuplePropManager(self.obj, attr)
    def CreateCollapsedVectorManager(self, attr):
        return TuplePropManager(self.obj, attr, suffix="XYZ", collaps=True)

    managerDict['name']         = CreateStringManager
    managerDict['__class__']    = CreateClassManager
    managerDict['gruppe']       = CreateGroupManager
    managerDict['sichtbar']     = CreateBoolManager

    managerDict['position']     = CreateVectorManager
    managerDict['laenge']       = CreateFloatManager
    managerDict['breite']       = managerDict['laenge']
    managerDict['hoehe']        = managerDict['laenge']
    managerDict['radius']       = managerDict['laenge']
    managerDict['achse']        = CreateCollapsedVectorManager
    managerDict['laengePfeil']  = CreateFloatManager
    managerDict['breitePfeil']  = managerDict['laengePfeil']
    managerDict['breiteSpitze'] = managerDict['laengePfeil']
    managerDict['laengeSpitze'] = managerDict['laengePfeil']
    managerDict['dicke']        = CreateFloatManager
    managerDict['farbe']        = CreateColourManager
    managerDict['deckkraft']    = CreateFloatManager
    managerDict['material']     = CreateMaterialManager

    def CleanUpdate(self, newObj):
        '''Remove current properties and categories and update self.obj
        reference'''
        self.Clear()
        self.managerMap.clear()
        self.catGroup = None
        self.catData = None
        self.catAppearance = None
        
        # update object
        self.OnObjectChange(newObj)
        self.obj = newObj

    def SetObject(self, targetObject):
        '''Change displayed object'''
        if self.obj is targetObject: return # abort

        # remove all properties and categories
        self.CleanUpdate(targetObject)
        
        for a in filterAttributes(self.obj) :
            # handle categories
            if a in objectName and self.catGroup == None:
                self.catGroup = wxpg.PropertyCategory("Objekt")
                self.Append(self.catGroup)
            if a in objectAttributes and self.catData == None:
                self.catData = wxpg.PropertyCategory("Attribute")
                self.Append(self.catData)
                
            # append property
            manager = ObjectPG.managerDict[a](self, a)
            if manager: self.AppendManager(manager)

        if self.obj.__class__ is Gruppe:
            self.AppendObjectList("Enthaltene Objekte", self.obj.Liste)

    def AppendObjectList(self, msg, objs):
        self.Append(wxpg.PropertyCategory(msg))
        for o in objs:
            self.AppendManager(ObjectPropManager(o))

    def AppendManager(self, man, parentID=None):
        prop = man.GetProperty()
        if parentID:
            propID = self.AppendIn(parentID, prop)
        else:
            propID = self.Append(prop)
        self.managerMap[prop.GetName()] = man
        man.PostInit(self, propID)

    def Update(self):
        '''checks all properties if their values changed'''
        for prop in self.Properties:
            self.managerMap[prop.GetName()].Update()

    def OnValueChange(self, event):
        prop = event.GetProperty()
        # do nothing if event did not have associated property
        if prop is None: return

        pendingValue = event.GetValue()        
        # handle the case where property value is unspecified
        if pendingValue is None: return
        
        # handle changes in values, as needed
        upd = self.managerMap[prop.GetName()].Process(pendingValue)
        if upd: self.Update()

    def GetObject(self):
        return self.obj
